(* absyn/absyn.sig *)

signature ABSYN =
  sig

    structure Source	: SOURCE

    (* start line, start column, end line, endcolumn *)
	datatype loc = LOC of int * int * int * int

    (* filename, pos start, pos end *)
	datatype info = INFO of string * int * int * loc ref

    datatype ident	= IDENT of string * info ref
    type var		= ident
    type tyvar		= ident

    datatype longid	= LONGID of ident option * ident * info ref

    datatype lit = CCONlit of char * info ref
				 | ICONlit of int * info ref
				 | RCONlit of real * info ref
				 | SCONlit of string * info ref

    datatype ty	= VARty of tyvar * info ref
				| CONSty of ty list * longid * info ref
				| TUPLEty of ty list * info ref
				| RELty of ty list * ty list * info ref

    datatype pat = WILDpat of info ref
				 | LITpat of lit * info ref
				 | CONpat of longid * info ref
				 | STRUCTpat of longid option * pat list * info ref
				 | BINDpat of var * pat * info ref
				 | IDENTpat of ident * pat ref * info ref

    datatype exp = LITexp of lit * info ref
				 | CONexp of longid * info ref
				 | VARexp of longid * info ref
				 | STRUCTexp of longid option * exp list * info ref
				 | IDENTexp of longid * exp ref * info ref

    datatype goal = CALLgoal of longid * exp list * pat list * info ref
				  | EQUALgoal of var * exp * info ref
				  | LETgoal of pat * exp * info ref
				  | NOTgoal of goal * info ref
				  | ANDgoal of goal * goal * info ref

    datatype result	= RETURN of exp list * info ref
					| FAIL of info ref
					
    datatype clause	= CLAUSE1 of goal option * ident * pat list * result * info ref
					| CLAUSE2 of clause * clause * info ref 

    datatype conbind = CONcb of ident * info ref
					 | CTORcb of ident * ty list * info ref

    datatype datbind = DATBIND of tyvar list * ident * conbind list * info ref

    datatype typbind = TYPBIND of tyvar list * ident * ty * info ref

    datatype relbind = RELBIND of ident * ty option * clause * info ref

    datatype spec = WITHspec of string * interface ref * info ref
				  | ABSTYPEspec of bool * tyvar list * ident * info ref
				  | TYPEspec of typbind list * info ref
				  | DATAspec of datbind list * typbind list * info ref
				  | VALspec of ident * ty * info ref
				  | RELspec of ident * ty * info ref

    and interface = INTERFACE of {modid: ident,
					specs: spec list,
					source: Source.source} * info ref

    datatype dec = WITHdec of string * interface ref * info ref
			| TYPEdec of typbind list * info ref
			| DATAdec of datbind list * typbind list * info ref
			| VALdec of ident * exp * info ref
			| RELdec of relbind list * info ref

    datatype module	= MODULE of interface * dec list * info ref

    structure IdentDict : ORD_DICT where type Key.ord_key = ident
    val makeIdent	    : string * info ref -> ident
    val rmlIdent	    : string -> ident
    val identName	    : ident -> string
    (* 2004-11-28 adrpo added for getting the info out of an ident *)
    val identCtxInfo	: ident -> info
    val lidentName		: longid -> string
    val lidentCtxInfo	: longid -> info
    val identEqual	    : ident * ident -> bool
    val litEqual	    : lit * lit -> bool
    val litString	    : lit -> string
    val dummyLoc        : loc
	val dummyInfo       : info
    val dummyInterface	: interface	
	
  end (* signature ABSYN *)
