/* parsutil.c */
#include <stdio.h>
#include "rml.h"
#include "yacclib.h"
#include "absyn.h"
#include "parsutil.h"

rml_value_t* pu_Constant_INTcon(rml_value_t* icon)		/* Absyn.INTcon=BOX1 */
{
    return mk_box1(Absyn__INTcon_3dBOX1, icon);
}

rml_value_t* pu_Constant_REALcon(rml_value_t* rcon)		/* Absyn.REALcon=BOX1 */
{
    return mk_box1(Absyn__REALcon_3dBOX1, rcon);
}

rml_value_t* pu_Constant_IDENTcon(rml_value_t* id)		/* Absyn.IDENTcon=BOX1 */
{
    return mk_box1(Absyn__IDENTcon_3dBOX1, id);
}

rml_value_t* pu_CONBND(rml_value_t* id, rml_value_t* con)	/* Absyn.CONBND=BOX2 */
{
    return mk_box2(Absyn__CONBND_3dBOX2, id, con);
}

rml_value_t* pu_Ty_NAME(rml_value_t* id)	/* Absyn.NAME=BOX1 */
{
    return mk_box1(Absyn__NAME_3dBOX1, id);
}

rml_value_t* pu_Ty_PTR(rml_value_t* ty)	/* Absyn.PTR=BOX1 */
{
    return mk_box1(Absyn__PTR_3dBOX1, ty);
}

rml_value_t* pu_Ty_ARR(rml_value_t* con, rml_value_t* ty)	/* Absyn.ARR=BOX2 */
{
    return mk_box2(Absyn__ARR_3dBOX2, con, ty);
}

rml_value_t* pu_Ty_REC(rml_value_t* varbnds)		/* Absyn.REC=BOX1 */
{
    return mk_box1(Absyn__REC_3dBOX1, varbnds);
}

rml_value_t* pu_VARBND(rml_value_t* id, rml_value_t* ty)	/* Absyn.VARBND=BOX2 */
{
    return mk_box2(Absyn__VARBND_3dBOX2, id, ty);
}

rml_value_t* pu_TYBND(rml_value_t* id, rml_value_t* ty)	/* Absyn.TYBND=BOX2 */
{
    return mk_box2(Absyn__TYBND_3dBOX2, id, ty);
}

rml_value_t* pu_Exp_INT(rml_value_t* icon)	/* Absyn.INT=BOX1 */
{
    return mk_box1(Absyn__INT_3dBOX1, icon);
}

rml_value_t* pu_Exp_REAL(rml_value_t* rcon)	/* Absyn.REAL=BOX1 */
{
    return mk_box1(Absyn__REAL_3dBOX1, rcon);
}

rml_value_t* pu_Exp_IDENT(rml_value_t* id)	/* Absyn.IDENT=BOX1 */
{
    return mk_box1(Absyn__IDENT_3dBOX1, id);
}

rml_value_t* pu_Exp_CAST(rml_value_t* ty, rml_value_t* exp)	/* Absyn.CAST=BOX2 */
{
    return mk_box2(Absyn__CAST_3dBOX2, ty, exp);
}

rml_value_t* pu_Exp_FIELD(rml_value_t* exp, rml_value_t* id)	/* Absyn.FIELD=BOX2 */
{
    return mk_box2(Absyn__FIELD_3dBOX2, exp, id);
}

rml_value_t* pu_Exp_UNARY(enum uop uop, rml_value_t* exp)	/* Absyn.UNARY=BOX2 */
{
    rml_uint_t unop;
    switch( uop ) {
      case UOP_ADDR:	unop = Absyn__ADDR_3dINT; break;
      case UOP_INDIR:	unop = Absyn__INDIR_3dINT; break;
      case UOP_NOT:		unop = Absyn__NOT_3dINT; break;
      case UOP_PLUS:	return exp;
      case UOP_MINUS:	return pu_Exp_BINARY(pu_Exp_INT(mk_icon(0)), BOP_SUB, exp);
    }
    return mk_box2(Absyn__UNARY_3dBOX2,
		    mk_icon(unop),
		    exp);
}

rml_value_t* pu_Exp_BINARY(rml_value_t* exp1, enum bop bop, rml_value_t* exp2) /* Absyn.BINARY=BOX3 */
{
    rml_uint_t binop;
    switch( bop ) 
	{
      case BOP_ADD:		binop = Absyn__ADD_3dINT; break;
      case BOP_SUB:		binop = Absyn__SUB_3dINT; break;
      case BOP_MUL:		binop = Absyn__MUL_3dINT; break;
      case BOP_RDIV:	binop = Absyn__RDIV_3dINT; break;
      case BOP_IDIV:	binop = Absyn__IDIV_3dINT; break;
      case BOP_IMOD:	binop = Absyn__IMOD_3dINT; break;
      case BOP_IAND:	binop = Absyn__IAND_3dINT; break;
      case BOP_IOR:	binop = Absyn__IOR_3dINT; break;
    }
    return mk_box3(Absyn__BINARY_3dBOX3,
		    exp1,
		    mk_icon(binop),
		    exp2);
}

rml_value_t* pu_Exp_RELATION(rml_value_t* exp1, enum rop rop, rml_value_t* exp2) /* Absyn.RELATION=BOX3 */
{
    rml_value_t* lhs, *rhs;
    rml_uint_t relop;
    switch( rop ) 
	{
      case ROP_LT:	relop = Absyn__LT_3dINT; lhs = exp1; rhs = exp2; break;
      case ROP_LE:	relop = Absyn__LE_3dINT; lhs = exp1; rhs = exp2; break;
      case ROP_GE:	relop = Absyn__LE_3dINT; lhs = exp2; rhs = exp1; break;
      case ROP_GT:	relop = Absyn__LT_3dINT; lhs = exp2; rhs = exp1; break;
    }
    return mk_box3(Absyn__RELATION_3dBOX3,
		    lhs,
		    mk_icon(relop),
		    rhs);
}

rml_value_t* pu_Exp_EQUALITY(rml_value_t* exp1, enum eop eop, rml_value_t* exp2) /* Absyn.EQUALITY=BOX2 */
{
    rml_value_t* exp = mk_box2(Absyn__EQUALITY_3dBOX2, exp1, exp2);
    if( eop == EOP_NE )
	exp = pu_Exp_UNARY(UOP_NOT, exp);
    return exp;
}

rml_value_t* pu_Exp_FCALL(rml_value_t* id, rml_value_t* args)	/* Absyn.FCALL=BOX2 */
{
    return mk_box2(Absyn__FCALL_3dBOX2, id, args);
}

rml_value_t* pu_Stmt_ASSIGN(rml_value_t* lhs, rml_value_t* rhs)	/* Absyn.ASSIGN=BOX2 */
{
    return mk_box2(Absyn__ASSIGN_3dBOX2, lhs, rhs);
}

rml_value_t* pu_Stmt_PCALL(rml_value_t* id, rml_value_t* args)	/* Absyn.PCALL=BOX2 */
{
    return mk_box2(Absyn__PCALL_3dBOX2, id, args);
}

rml_value_t* pu_Stmt_FRETURN(rml_value_t* exp)	/* Absyn.FRETURN=BOX1 */
{
    return mk_box1(Absyn__FRETURN_3dBOX1, exp);
}

rml_value_t* pu_Stmt_PRETURN(void)	/* Absyn.PRETURN=BOX0 */
{
    return mk_box0(Absyn__PRETURN_3dBOX0);
}

rml_value_t* pu_Stmt_WHILE(rml_value_t* exp, rml_value_t* stmt)	/* Absyn.WHILE=BOX2 */
{
    return mk_box2(Absyn__WHILE_3dBOX2, exp, stmt);
}

rml_value_t* pu_Stmt_IF(rml_value_t* exp, rml_value_t* stmt1, rml_value_t* stmt2)	/* Absyn.IF=BOX3 */
{
    return mk_box3(Absyn__IF_3dBOX3, exp, stmt1, stmt2);
}

rml_value_t* pu_Stmt_SEQ(rml_value_t* stmt1, rml_value_t* stmt2)	/* Absyn.SEQ=BOX2 */
{
    return mk_box2(Absyn__SEQ_3dBOX2, stmt1, stmt2);
}

rml_value_t* pu_Stmt_SKIP(void)	/* Absyn.SKIP=BOX0 */
{
    return mk_box0(Absyn__SKIP_3dBOX0);
}

rml_value_t* pu_SubBnd_FUNCBND(rml_value_t* id, rml_value_t* varbnds, rml_value_t* ty, rml_value_t* block_opt) /* Absyn.FUNCBND=BOX4 */
{
    return mk_box4(Absyn__FUNCBND_3dBOX4, id, varbnds, ty, block_opt);
}

rml_value_t* pu_SubBnd_PROCBND(rml_value_t* id, rml_value_t* varbnds, rml_value_t* block_opt)	/* Absyn.PROCBND=BOX3 */
{
    return mk_box3(Absyn__PROCBND_3dBOX3, id, varbnds, block_opt);
}

rml_value_t* pu_BLOCK(rml_value_t* conbnds, rml_value_t* tybnds, rml_value_t* varbnds, rml_value_t* subbnds, rml_value_t* stmt)	/* Absyn.BLOCK=BOX5 */
{
    return mk_box5(Absyn__BLOCK_3dBOX5, conbnds, tybnds, varbnds, subbnds, stmt);
}

rml_value_t* pu_PROG(rml_value_t* id, rml_value_t* block)	/* Absyn.PROG=BOX2 */
{
    return mk_box2(Absyn__PROG_3dBOX2, id, block);
}
