/* misc_print.c */
#include <stdio.h>
#include "rml.h"

RML_BEGIN_LABEL(RML__debug)
{
	/*
	   * A0=fileName:string
	   * A1=sline:int, A2=scolumn:int, A3=eline:int, A4=ecolumn:int
	   * A5=relation:string
	   * A6=goal:string
	*/
    int i;
    void *str = rmlA0;
    void *strrel = rmlA5; void *strgoal = rmlA6; 
    char s[RML_DEBUG_MAX_STRING] = ""; 
    snprintf(s, RML_DEBUG_MAX_STRING, "->%s:%d.%d.%d.%d relation[%s].goal[%s]", 
				RML_STRINGDATA(str),		/* file */ 
				RML_UNTAGFIXNUM(rmlA1),		/* start line */
				RML_UNTAGFIXNUM(rmlA2),		/* start column */
				RML_UNTAGFIXNUM(rmlA3),		/* end line */
				RML_UNTAGFIXNUM(rmlA4),     /* end column */
				RML_STRINGDATA(strrel),     /* relation */
				RML_STRINGDATA(strgoal)     /* goal (term) */
				);	
    /* add the node in stack */
    if (execution_type != RMLDB_FAST) push_node(s);
    if (number_of_filters != 0)
    {
        for (i = 0; i < number_of_filters; i++)
            if (strstr(s, relation_filter[i]) != NULL)
            {
                printf("\nBreakpoint [%d], on %s reached", i, relation_filter[i]);
                execution_type = RMLDB_STEP;
                rmldb_show = RMLDB_SHOW;
                break;
            }
    }
	/* printf("%s\n", s); */
    /* test if we are stepping */
    if (execution_type == RMLDB_STEP) /* do we step? */
    {
		char b[RML_DEBUG_MAX_STRING] = "";
		/* updating the execution location -- needed for type reconstruction */
		sprintf(rml_current_execution_loc.file,"%s", RML_STRINGDATA(str)); 
        sprintf(rml_current_execution_loc.relation, "%s", RML_STRINGDATA(strrel));
        sprintf(rml_current_execution_loc.goal, "%s", RML_STRINGDATA(strgoal));
		rml_current_execution_loc.range.sl = RML_UNTAGFIXNUM(rmlA1),	/* start line */
		rml_current_execution_loc.range.sc = RML_UNTAGFIXNUM(rmlA2),	/* start column */
		rml_current_execution_loc.range.el = RML_UNTAGFIXNUM(rmlA3),	/* end line */
		rml_current_execution_loc.range.ec = RML_UNTAGFIXNUM(rmlA4),   /* end column */

        rmldb_sprintf ("\n%s", RML_STRINGDATA(str));    
        rmldb_sprintf (":"); rmldb_sprintf ("%d", RML_UNTAGFIXNUM(rmlA1));
        rmldb_sprintf ("."); rmldb_sprintf ("%d", RML_UNTAGFIXNUM(rmlA2));
        rmldb_sprintf ("@%s", RML_STRINGDATA(strrel));
        rmldb_sprintf ("@%s", RML_STRINGDATA(strgoal));
        /* this is for emacs-mode */
        snprintf(b, RML_DEBUG_MAX_STRING, "\n@%s:%d.%d.%d.%d relation[%s].goal[%s]\n", 
				RML_STRINGDATA(str),		/* file */ 
				RML_UNTAGFIXNUM(rmlA1),		/* start line */
				RML_UNTAGFIXNUM(rmlA2),		/* start column */
				RML_UNTAGFIXNUM(rmlA3),		/* end line */
				RML_UNTAGFIXNUM(rmlA4),     /* end column */
				RML_STRINGDATA(strrel),     /* relation */
				RML_STRINGDATA(strgoal)     /* goal (term) */
				);
        rmldb_sprintf("%s", b);
		if (rml_debug_print_livevars_each_step)
			print_rml_debug_vars();

		print_displayvars();
        /* parse a command */
        rmldb_parse();
    }
	if (execution_type != RMLDB_FAST) clear_rml_debug_vars();

    RML_TAILCALLK(rmlSC);
}
RML_END_LABEL
