/**************************************************

 [ rml-debug.h ]

 adrpo 2002-10
 interface for rml debugging 

 this file contains all the functions for debugging
 
***************************************************/

#define _RML_DEBUG_

#ifndef _RML_DEBUG_H_
#define _RML_DEBUG_H_

#include <stdio.h>

#define RMLDB_PROMPT "rmldb@>"

#define rmldb_sprintf printf

/* adrpo 
   some useful defines. */
#define RMLDB_RUN   0
#define RMLDB_STEP  1
#define RMLDB_FAST  3 /* no livevars, no backtrace, just very fast debugging (breakpoints) */

#define RMLDB_SHOW  1
#define RMLDB_HIDE  0

#define RML_DEBUG_VAR_IN 0
#define RML_DEBUG_VAR_OUT 1

#define RML_LIVEVARS_ON  1
#define RML_LIVEVARS_OFF 0

#define RML_DEBUG_MAX_BREAKPOINTS 100
#define RML_DEBUG_MAX_STRING  1000

#define RML_DEBUG_MAX_DISPLAY_VARS 100

/* adrpo 
   this flag is set to RMLDB_STEP when you need a "step" by step execution 
   if you want "run" execution then this flag is set on RMLDB_RUN */ 
extern int execution_type; 
extern int rmldb_show;
extern int last_command; /* need to know if we should start running or get more commands from the debugger */
extern char* rmldb_command;
extern int max_backtrace_entries;
extern int depth_of_variable_print;
extern int max_string_variable_print;
extern int rml_debug_print_livevars_each_step; /* print the current livevars each step */

extern void rml_var_print(void *p);
void show_help(void);
int quit(char *line);

/*
  adrpo 2002-10 creation
  - this function tries to connect through sockets to 
    the rmld process for debugging.
  - it will listen to the first command a user will 
    give
  - if an error in communication occurs -1 is returned
*/
extern int rml_debug_init(void);

/*
  adrpo 2002-10 creation
  - this function will end the socket communication
  - if an error in communication occurs -1 is returned
*/
extern int rml_debug_end(void);

/* adrpo 
   this filter is set on the relations you want to make the 
   program to stop (like a breakpoint, but with string matching)  */
extern char relation_filter[RML_DEBUG_MAX_BREAKPOINTS][RML_DEBUG_MAX_STRING];
extern int number_of_filters; /* of course, the number of filters. */

extern void add_relation_filter(char *);
extern void del_relation_filter(char *);

struct call_stack_node
{
   char relation_name[RML_DEBUG_MAX_STRING];
   int depth;
   struct call_stack_node *prev;
   struct call_stack_node *next;
};

struct rml_debug_var_node
{
   void* var_name;
   void* var;
   int depth;
   struct rml_debug_var_node *prev;
   struct rml_debug_var_node *next;
};

extern void rml_debug_add_var(int direction, void* var_name, void* var);
extern void clear_rml_debug_vars(void);
extern void rml_print_variable(char* var_name);
extern void rml_print_sizeof_variable(char* var_name);
extern void print_rml_debug_vars(void);
extern void set_print_rml_debug_vars(int flag);
extern char rml_display_vars[RML_DEBUG_MAX_DISPLAY_VARS][RML_DEBUG_MAX_STRING];
extern int number_of_display_vars; /* of course, the number active display vars. */
extern void rml_display_variable(char* var_name);
extern void rml_undisplay_variable(char* var_name);
extern void print_displayvars(void);
extern void show_displayvars(void);
extern void clear_displayvars(void);
extern void rml_view_variable(char* var_name);

extern char *rmldb_ttyname;

extern void push_node(char *rname);
extern void pop_node(void);
extern void print_backtrace(char* filter);
extern void rml_set_depth(char* depth);
extern void rml_set_maxstr(char* maxstr);
extern void show_breakpoints(void);
extern void clear_breakpoints(void);
extern void set_max_backtrace(char *maxbt);
extern void print_settings(void);
extern void set_rmldb_output(char *);

typedef struct rml_range_db
{
  int sl;
  int sc;
  int el;
  int ec;
} rml_range_db_t;

typedef struct rml_ex_loc
{
   char file[RML_DEBUG_MAX_STRING];
   rml_range_db_t range;
   char relation[RML_DEBUG_MAX_STRING];
   char goal[RML_DEBUG_MAX_STRING];
} rml_current_execution_loc_t;

/* location of current execution point */
extern rml_current_execution_loc_t rml_current_execution_loc;

/* types */

typedef struct rml_str
{
   char *name;
   int depth;
   struct rml_str *prev;
   struct rml_str *next;
} rml_str_t;

typedef struct rml_str_list
{
	rml_str_t* list_start;
	rml_str_t* list_end;
} rml_str_list_t;

/*
    excerpt from rml.grm
    datatype ident	= IDENT of string
    type tyvar		= ident
    datatype longid	= LONGID of ident option * ident
    datatype ty	= VARty of tyvar
				| CONSty of ty list * longid
				| TUPLEty of ty list
				| RELty of ty list * ty list
%nonterm ty of Absyn.ty
	   | tuple_ty of Absyn.ty list
	   | ty_sans_star of Absyn.ty
	   | ty_comma_seq2 of Absyn.ty list
	   | seq_ty of Absyn.ty list
*/

/* define type ahead, as is it mutually recursive
 * with type components 
 */
struct rml_type;

typedef struct rmlLISTty
{
   int length;
   struct rml_type* list_start; 
   struct rml_type* list_end;
} rmlLISTty_t;

typedef struct rmlVARty { char* id; } rmlVARty_t;

typedef struct rmlCONSty 
{ 
	rmlLISTty_t* list;
	char* id;
} rmlCONSty_t;

typedef struct rmlTUPLEty 
{ 
	rmlLISTty_t* list;
} rmlTUPLEty_t;

typedef struct rmlRELty 
{ 
	rmlLISTty_t* list1;
	rmlLISTty_t* list2;
} rmlRELty_t;


typedef enum tyKind {eNORMAL,eLISTty,eVARty,eCONSty,eTUPLEty,eRELty} tyKind_t;

typedef struct rml_type
{
   tyKind_t kind; /* selects the stufs in the union or no union at all */
   union component_t
   {
	 rmlLISTty_t*  l;
	 rmlVARty_t*   v;
	 rmlCONSty_t*  c;
	 rmlTUPLEty_t* t;
	 rmlRELty_t*   r;
   } component;

   /* these are used in rmlLISTty */
   int depth;
   struct rml_type *prev;
   struct rml_type *next;
} rml_type_t;

typedef struct rml_type_db
{
   char *file;
   char *name;
   rml_range_db_t* range;
   rml_type_t *type;
   int depth;
   struct rml_type_db *prev;
   struct rml_type_db *next;
} rml_type_db_t;

typedef struct rml_var_db
{
   char *file;
   rml_range_db_t *range;
   rml_range_db_t *clause_range;
   char *relation;
   char *name;
   rml_type_db_t* type_db;

   int depth;
   struct rml_var_db *prev;
   struct rml_var_db *next;
} rml_var_db_t;

typedef struct rml_con_db
{
   char *file;
   int constructor;
   int is_transparent;
   rml_range_db_t* range;
   char *name;
   rml_type_db_t* type_db;

   int depth;
   struct rml_con_db *prev;
   struct rml_con_db *next;
} rml_con_db_t;

typedef struct rml_relation_db
{
   char *file;
   rml_range_db_t* range;
   char *name;

   rml_type_db_t* type_db;
	
   int depth;
   struct rml_relation_db *prev;
   struct rml_relation_db *next;
} rml_relation_db_t;

extern rml_var_db_t *rml_var_db_start;
extern rml_var_db_t *rml_var_db_end;
extern rml_type_db_t *rml_type_db_start;
extern rml_type_db_t *rml_type_db_end;
extern rml_con_db_t *rml_con_db_start;
extern rml_con_db_t *rml_con_db_end;
extern rml_relation_db_t *rml_relation_db_start;
extern rml_relation_db_t *rml_relation_db_end;

extern int rmldb_load_db(char* programdb_file);
extern void rml_print_type(rml_type_t *type, FILE* out);
extern void rml_print_type_list(rmlLISTty_t* list, char* separator, FILE* out);
extern rml_type_db_t* rmldb_get_type(char* var_name);
extern char* rml_get_con_name(rml_type_t *type, int constructor);

extern void rml_str_list_free(rml_str_list_t* list);
extern void rml_type_list_to_list(rmlLISTty_t* list, rml_str_list_t* lst);
extern void rml_type_to_list(rml_type_t *type, rml_str_list_t* list);
extern int rml_str_list_get_size(rml_str_list_t* list);
extern char* rml_str_list_get_index(rml_str_list_t* list, int index);
extern rmlLISTty_t* rml_get_type_components(rml_con_db_t* con_db);
extern rml_type_t* rml_get_type_component(rml_type_t* type, int constr, int index);
extern rml_type_t* rml_get_type_comp_from_con(rmlLISTty_t* type, int index);
extern void rml_var_show(void *p, rml_type_t* type, int depth);

/* ---- components ---- */
/* create a list from two tys */
extern rmlLISTty_t* make_rmlLISTty(rml_type_t* type1, rml_type_t* type2);
/* appends ty in front of the list */
extern rmlLISTty_t* make_cons_rmlLISTty(rml_type_t* type, rmlLISTty_t* list);
/* make a VARty */
extern rmlVARty_t* make_rmlVARty(char* id);
/* make a CONSty */
extern rmlCONSty_t* make_rmlCONSty(rmlLISTty_t* list, char* id);
/* make a TUPLEty */
extern rmlTUPLEty_t* make_rmlTUPLEty(rmlLISTty_t* list);
/* make a RELty */
extern rmlRELty_t* make_rmlRELty(rmlLISTty_t* list1, rmlLISTty_t* list2);

/* makes a type from a component, and casts it according to kind */
extern rml_type_t* make_rml_type(tyKind_t kind, void* component);

extern rml_str_t* make_rml_str(char* name);

extern rml_var_db_t* make_rml_var_db(
		char* file,
		rml_range_db_t* range,
		rml_range_db_t* clause_range,
		char* relation,
		char* name,
		rml_type_db_t* type_db);

extern rml_con_db_t* make_rml_con_db(
		char* file,
		rml_range_db_t* range,
		char* name,
		rml_type_db_t* type_db,
		int constructor);

extern rml_type_db_t* make_rml_type_db(
		char* file,
		rml_range_db_t* range,
		char* name);

extern rml_type_db_t* make_rml_type_db_t(
		rml_type_t *type);

extern rml_relation_db_t* make_rml_relation_db(
		char* file,
		rml_range_db_t* range,
		char* name,
		rml_type_db_t* type_db);

extern void rml_type_list_push_front(
					rml_type_t **start_node, 
				    rml_type_t **end_node, 
					rml_type_t *node);

extern void rml_type_list_push_back(
					rml_type_t **start_node, 
				    rml_type_t **end_node, 
					rml_type_t *node);

extern void rml_str_list_add(
					rml_str_t **start_node, 
				    rml_str_t **end_node, 
					rml_str_t *node);

extern void rml_var_db_add(
					rml_var_db_t **start_node, 
				    rml_var_db_t **end_node, 
					rml_var_db_t *node);
extern void rml_con_db_add(
					rml_con_db_t **start_node, 
				    rml_con_db_t **end_node, 
					rml_con_db_t *node);
extern void rml_type_db_add(
					rml_type_db_t **start_node, 
				    rml_type_db_t **end_node, 
					rml_type_db_t *node);
extern void rml_relation_db_add(
					rml_relation_db_t **start_node, 
				    rml_relation_db_t **end_node, 
					rml_relation_db_t *node);

/* program database stream */
extern FILE *aarmldbin; /* the stream we need to parse from the rmldebugger */
extern int aarmldberror(char*);
extern int aarmldbparse(void);
extern int aarmldbdebug;


/* comand stream */
extern FILE *aain; /* the stream we need to parse from the rmldebugger */
extern int aaerror(char*);
extern int aaparse(void);
extern int aadebug;

int open_new_file(void);
void use_quit (int);

int rmldb_parse(void);

#endif /* [ rml-debug.h]  */
