%{
    
#include "rml-debug.h"

extern int aalex(void);

#define YYDEBUG 1
#define YYERROR_VERBOSE                  /* Have this defined to give better
                                            error messages. Using it causes
					    some bison warnings at compiler
					    compile time, however. Use as you
					    wish. Not mandatory. */
%}


/* The different semantic values that can be returned within the AST. 
 */
%union 
{
  int   v_int;
  float v_float;
  char *to_print;
  char *id;
}

/* uncomment this shit on linux and comment it on solaris
   damn incompatibilities
*/
/* %name-prefix="aa" */

/* Here are return types for all productions that return nodes (ie, the
   return type of the bison $$ construct).
   (hint: maybe $$ has the type YYSTYPE?). */
%token T_EOF T_ERROR T_EOL T_HELP T_MAXBT
%token T_BREAKPOINT T_SET T_SETTINGS T_RUN T_STEP T_CLEAR T_FILTERED_BACKTRACE
%token T_BACKTRACE T_QUIT T_ON T_OFF T_DEPTH T_SHOW T_STATUS T_END T_DEBUGPARSER
%token T_OUTPUT T_LIVEVARS T_PRINT T_SIZEOF T_DISPLAY T_UNDISPLAY T_GRAPH T_FAST
%token T_MAXSTR
%token <char *>T_ID

%start command
%%

command:  
T_BREAKPOINT T_ON T_ID T_EOL
{
  /* set On the filter on this relation */
  add_relation_filter(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_BREAKPOINT T_OFF T_ID T_EOL
{
  /* set Off the filter on this relation */
  del_relation_filter(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_SET T_STEP T_ON T_EOL
{
  execution_type = RMLDB_STEP;
  last_command = 0;
  YYACCEPT;
}
| T_SET T_STEP T_OFF T_EOL
{
  execution_type = RMLDB_RUN;
  last_command = 0;
  YYACCEPT;
}
| T_STEP T_EOL
{
  last_command = 1;
  YYACCEPT;
}
| T_RUN T_EOL
{
  execution_type = RMLDB_RUN;
  last_command = 1;
  rmldb_show = RMLDB_HIDE;
  YYACCEPT;
}
| T_BACKTRACE T_EOL
{
  print_backtrace(NULL);
  last_command = 0;
  YYACCEPT;
}
| T_FILTERED_BACKTRACE T_ID T_EOL
{
  print_backtrace(aalval.id);
  last_command = 0;
  YYACCEPT;   
}
| T_HELP T_EOL
{
  show_help(); 
  last_command = 0;
  YYACCEPT;
}
| T_QUIT T_EOL
{
  printf("end the program\n");
  last_command = 0;
  rml_exit(2);
}
| T_SET T_DEPTH T_ID T_EOL
{
  rml_set_depth(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_DEPTH T_ID T_EOL
{
  rml_set_depth(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_SET T_MAXSTR T_ID T_EOL
{
  rml_set_maxstr(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_MAXSTR T_ID T_EOL
{
  rml_set_maxstr(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_SET T_MAXBT T_ID T_EOL
{
  set_max_backtrace(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_MAXBT T_ID T_EOL
{
  set_max_backtrace(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_SETTINGS T_EOL
{
  print_settings();
  last_command = 0;
  YYACCEPT;
}
| T_SHOW T_EOL
{
  show_breakpoints();
  last_command = 0;
  YYACCEPT;
}
| T_CLEAR T_EOL
{
  clear_breakpoints();
  last_command = 0;
  YYACCEPT;
}
| T_STATUS T_EOL
{
  rml_show_status();
  last_command = 0;
  YYACCEPT;
}
| T_OUTPUT T_ID T_EOL
{
  set_rmldb_output(aalval.id);
  last_command = 0;
  YYACCEPT;
}
| T_PRINT T_ID T_EOL
{
  last_command = 0;
  rml_print_variable(aalval.id);
  YYACCEPT;
}
| T_SIZEOF T_ID T_EOL
{
  last_command = 0;
  rml_print_sizeof_variable(aalval.id);
  YYACCEPT;
}
| T_LIVEVARS T_EOL
{
  last_command = 0;
  print_rml_debug_vars();
  YYACCEPT;
}
| T_SET T_LIVEVARS T_ON T_EOL
{
  last_command = 0;
  set_print_rml_debug_vars(RML_LIVEVARS_ON);
  YYACCEPT;
}
| T_LIVEVARS T_ON T_EOL
{
  last_command = 0;
  set_print_rml_debug_vars(RML_LIVEVARS_ON);
  YYACCEPT;
}
| T_SET T_LIVEVARS T_OFF T_EOL
{
  last_command = 0;
  set_print_rml_debug_vars(RML_LIVEVARS_OFF);
  YYACCEPT;
}
| T_LIVEVARS T_OFF T_EOL
{
  last_command = 0;
  set_print_rml_debug_vars(RML_LIVEVARS_OFF);
  YYACCEPT;
}
| T_DISPLAY T_EOL
{
  last_command = 0;
  show_displayvars();
  YYACCEPT;
}
| T_DISPLAY T_ID T_EOL
{
  last_command = 0;
  rml_display_variable(aalval.id);
  YYACCEPT;
}
| T_UNDISPLAY T_EOL
{
  last_command = 0;
  clear_displayvars();
  YYACCEPT;	
}
| T_UNDISPLAY T_ID T_EOL
{
  last_command = 0;
  rml_undisplay_variable(aalval.id);
  YYACCEPT;
}
| T_FAST T_EOL
{
  printf("Entering in fast debug mode, no backtrace generation, no live variables.\n");
  printf("Only verifying breakpoints, if they exist. \n");
  printf("Hit CTRL+C to enter again in Step execution mode when you want.\n");
  last_command = 1;
  execution_type = RMLDB_FAST;
  rmldb_show = RMLDB_HIDE;
  YYACCEPT;
}
| T_GRAPH T_ID T_EOL
{
  last_command = 0;
  rml_view_variable(aalval.id);
  YYACCEPT;
}
| T_DEBUGPARSER T_ON T_EOL
{
  aadebug = 1;
  YYACCEPT;
}
| T_DEBUGPARSER T_OFF T_EOL
{
  aadebug = 0;
  YYACCEPT;
}
;

%%

int aaerror(char *s) 
{
  fprintf(stderr, "%s or command not recognized\n", s);
  return -1;
}
